import base64
import logging
from datetime import datetime
from typing import Iterable, Tuple

import pytz

from fetchfox.apis.algorand import (
    algonodecloud,
    nfdomains,
    randswapcom,
)
from fetchfox.blockchains.base import Blockchain
from fetchfox.constants.blockchains import ALGORAND
from fetchfox.constants.currencies import ALGO
from fetchfox.constants.marketplaces import (
    RANDGALLERY_COM,
)
from fetchfox.dtos import (
    AssetDTO,
    CampaignDTO,
    FloorDTO,
    HoldingDTO,
    ListingDTO,
    SaleDTO,
    TransactionDTO,
    TransactionInputOutputDTO,
    TransactionAssetDTO,
)
from . import utils
from .exceptions import (
    InvalidAlgorandAssetIdException,
    InvalidAlgorandCollectionIdException,
    InvalidAlgorandAccountException,
)

logger = logging.getLogger(__name__)


class Algorand(Blockchain):
    def __init__(self):
        super().__init__(
            name=ALGORAND,
            currency=ALGO,
            logo="https://s2.coinmarketcap.com/static/img/coins/64x64/4030.png",
        )

    def check_account(self, account: str):
        if not utils.is_account(account):
            raise InvalidAlgorandAccountException(account)

    def check_asset_id(self, asset_id: str):
        if not utils.is_asset_id(asset_id):
            raise InvalidAlgorandAssetIdException(asset_id)

    def check_collection_id(self, collection_id: str):
        if not utils.is_address(collection_id):
            raise InvalidAlgorandCollectionIdException(collection_id)

    def explorer_url(self, *, address: str = None, collection_id: str = None, asset_id: str = None, tx_hash: str = None) -> str:
        if address:
            return f"https://algoexplorer.io/address/{address.upper()}"

        if asset_id:
            return f"https://algoexplorer.io/asset/{asset_id}"

        if collection_id:
            return f"https://algoexplorer.io/address/{collection_id.upper()}"

        if tx_hash:
            return f"https://algoexplorer.io/tx/{tx_hash.upper()}"

        return None

    def marketplace_url(self, *, collection_id: str = None, asset_id: str = None) -> str:
        if asset_id:
            return f"https://www.randgallery.com/algo-collection/?address={asset_id}"

        if collection_id:
            return f"https://randgallery.com/algo-collection/?address={collection_id}"

        return None

    # Accounts

    def get_account_assets(self, account: str, collection_id: str = None) -> Iterable[HoldingDTO]:
        self.check_account(account)

        if utils.is_nf_domain(account):
            account = nfdomains.resolve_nf_domain(account)

        response = algonodecloud.get_account_assets(account)

        for holding in response:
            asset_id = holding["asset-id"]
            quantity = holding["amount"]

            if quantity < 1:
                continue

            yield HoldingDTO(
                collection_id=None,
                asset_id=asset_id,
                address=account,
                quantity=quantity,
            )

    def get_account_balance(self, account: str) -> Tuple[float, str]:
        self.check_account(account)

        if utils.is_nf_domain(account):
            account = nfdomains.resolve_nf_domain(account)

        balance = algonodecloud.get_account_balance(account)

        return balance, self.currency

    def get_account_name(self, account: str) -> str:
        if utils.is_nf_domain(account):
            return account

        if utils.is_address(account):
            return nfdomains.get_nf_domain(account)

        return None

    def resolve_account_name(self, name: str) -> str:
        if utils.is_nf_domain(name):
            return nfdomains.resolve_nf_domain(name)

        return None

    # Assets

    def get_asset(self, collection_id: str, asset_id: str, fetch_metadata: bool = True, *args, **kwargs) -> AssetDTO:
        if collection_id:
            self.check_collection_id(collection_id)

        self.check_asset_id(asset_id)

        asset_data = algonodecloud.get_asset_data(asset_id)

        if fetch_metadata:
            metadata = algonodecloud.get_asset_metadata(asset_id)
            metadata["name"] = asset_data["name"]
        else:
            metadata = {}

        return AssetDTO(
            collection_id=asset_data["creator"],
            asset_id=asset_id,
            metadata=metadata,
        )

    def get_asset_owners(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        asset_owner = algonodecloud.get_asset_owner(str(asset_id))

        yield HoldingDTO(
            collection_id=collection_id,
            asset_id=asset_owner["asset_id"],
            address=asset_owner["address"],
            quantity=asset_owner["amount"],
        )

    def get_account_transactions(self, account: str, last: int = 10) -> Iterable[TransactionDTO]:
        self.check_account(account)

        if utils.is_nf_domain(account):
            account = nfdomains.resolve_nf_domain(account)

        for index, transaction in enumerate(algonodecloud.get_account_transactions(account)):
            if index >= last:
                break

            try:
                message = base64.b64decode(transaction["note"]).decode("utf-8")
            except:
                message = None

            sender = transaction["sender"]

            if transaction["tx-type"] == "pay":
                amount = transaction["payment-transaction"]["amount"]
                receiver = transaction["payment-transaction"]["receiver"]
                unit = "algos"
            elif transaction["tx-type"] == "axfer":
                amount = transaction["asset-transfer-transaction"]["amount"]
                receiver = transaction["asset-transfer-transaction"]["receiver"]
                unit = transaction["asset-transfer-transaction"]["asset-id"]
            else:
                continue

            yield TransactionDTO(
                blockchain=self.name,
                address=account,
                tx_hash=transaction["id"],
                message=message,
                inputs=[
                    TransactionInputOutputDTO(
                        address=sender,
                        assets=[
                            TransactionAssetDTO(
                                amount=amount,
                                unit=unit,
                            ),
                        ],
                    ),
                ],
                outputs=[
                    TransactionInputOutputDTO(
                        address=receiver,
                        assets=[
                            TransactionAssetDTO(
                                amount=amount,
                                unit=unit,
                            ),
                        ],
                    ),
                ],
            )

    # Collections

    def get_collection_assets(self, collection_id: str, fetch_metadata: bool = True, *args, **kwargs) -> Iterable[AssetDTO]:
        self.check_collection_id(collection_id)

        collection_assets = algonodecloud.get_collection_assets(collection_id)

        for asset_id in collection_assets:
            if fetch_metadata:
                yield self.get_asset(
                    collection_id=collection_id,
                    asset_id=str(asset_id),
                )
            else:
                yield AssetDTO(
                    collection_id=collection_id,
                    asset_id=str(asset_id),
                    metadata={},
                )

    def get_collection_floor(self, collection_id: str, *args, **kwargs) -> FloorDTO:
        self.check_collection_id(collection_id)

        collection_listings = self.get_collection_listings(collection_id)

        floor = None
        count = 0

        for listing in collection_listings:
            count += 1

            if floor is None:
                floor = listing
            elif listing.usd < floor.usd:
                floor = listing

        return FloorDTO(
            listing=floor,
            listing_count=count,
        )

    def get_collection_listings(self, collection_id: str, *args, **kwargs) -> Iterable[ListingDTO]:
        self.check_collection_id(collection_id)

        collection_listings = randswapcom.get_collection_listings(collection_id)

        for listing in collection_listings:
            asset_id = str(listing["assetId"])
            asset_ids = [asset_id]
            asset_names = [""]

            listed_at = datetime.fromtimestamp(
                listing["timestamp"] // 1000,
            ).replace(
                tzinfo=pytz.UTC,
            )

            yield ListingDTO(
                identifier=listing["timestamp"],
                collection_id=collection_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                listing_id=listing["timestamp"],
                marketplace=RANDGALLERY_COM,
                price=listing["price"],
                currency=ALGO,
                listed_at=listed_at,
                listed_by=listing["sellerAddress"],
                marketplace_url=f"https://randgallery.com/algo-collection/?address={asset_id}",
            )

    def get_collection_sales(self, collection_id: str, *args, **kwargs) -> Iterable[SaleDTO]:
        pass

    def get_collection_snapshot(self, collection_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)

        for asset in self.get_collection_assets(collection_id, fetch_metadata=False):
            yield from self.get_asset_owners(collection_id, asset.asset_id)

    # Campaigns

    def get_campaigns(self, starts_after: datetime = None) -> Iterable[CampaignDTO]:
        return



from abc import abstractmethod
from datetime import datetime
from typing import Iterable, Tuple

from fetchfox.apis import price
from fetchfox.dtos import (
    AssetDTO,
    CampaignDTO,
    FloorDTO,
    HoldingDTO,
    ListingDTO,
    RankDTO,
    SaleDTO,
    TransactionDTO,
)


class Blockchain:
    def __init__(self, name: str, currency: str, logo: str, preprod: bool = False):
        self.name: str = name
        self.currency: str = currency
        self.logo: str = logo
        self.preprod: bool = preprod

    @property
    def usd(self) -> float:
        return price.usd(self.currency)

    @abstractmethod
    def check_account(self, account: str):
        raise NotImplementedError()

    @abstractmethod
    def check_collection_id(self, collection_id: str):
        raise NotImplementedError()

    @abstractmethod
    def check_asset_id(self, asset_id: str):
        raise NotImplementedError()

    @abstractmethod
    def explorer_url(self, *, address: str = None, collection_id: str = None, asset_id: str = None, tx_hash: str = None) -> str:
        raise NotImplementedError()

    @abstractmethod
    def marketplace_url(self, *, collection_id: str = None, asset_id: str = None) -> str:
        raise NotImplementedError()

    # Accounts

    @abstractmethod
    def get_account_assets(self, wallet: str, collection_id: str = None) -> Iterable[HoldingDTO]:
        raise NotImplementedError()

    @abstractmethod
    def get_account_balance(self, account: str) -> Tuple[float, str]:
        raise NotImplementedError()

    @abstractmethod
    def get_account_transactions(self, account: str, last: int = 10) -> Iterable[TransactionDTO]:
        raise NotImplementedError()

    def get_account_last_transaction(self, account: str) -> TransactionDTO:
        try:
            return next(self.get_account_transactions(account, last=1))
        except StopIteration:
            return None

    @abstractmethod
    def get_account_name(self, account: str) -> str:
        raise NotImplementedError()

    @abstractmethod
    def resolve_account_name(self, account: str) -> str:
        raise NotImplementedError()

    def format_account(self, account: str) -> str:
        self.check_account(account)

        name = self.get_account_name(account)

        if name:
            return name

        return f"{account[:5]}..{account[-5:]}"

    # Assets

    @abstractmethod
    def get_asset(self, collection_id: str, asset_id: str, *args, **kwargs) -> AssetDTO:
        raise NotImplementedError()

    @abstractmethod
    def get_asset_owners(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        raise NotImplementedError()

    def get_asset_rank(self, collection_id: str, asset_id: str, *args, **kwargs) -> RankDTO:
        return None

    # Collections

    @abstractmethod
    def get_collection_assets(self, collection_id: str, fetch_metadata: bool = True, *args, **kwargs) -> Iterable[AssetDTO]:
        raise NotImplementedError()

    def get_collection_lock_date(self, collection_id: str, *args, **kwargs) -> datetime:
        return None

    @abstractmethod
    def get_collection_floor(self, collection_id: str, *args, **kwargs) -> FloorDTO:
        raise NotImplementedError()

    @abstractmethod
    def get_collection_listings(self, collection_id: str, *args, **kwargs) -> Iterable[ListingDTO]:
        raise NotImplementedError()

    def get_collection_ranks(self, collection_id: str, *args, **kwargs) -> Iterable[RankDTO]:
        return []

    @abstractmethod
    def get_collection_sales(self, collection_id: str, *args, **kwargs) -> Iterable[SaleDTO]:
        raise NotImplementedError()

    @abstractmethod
    def get_collection_snapshot(self, collection_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        raise NotImplementedError()

    # Campaigns

    def get_campaigns(self, starts_after: datetime = None) -> Iterable[CampaignDTO]:
        raise NotImplementedError()

    # Others
    def verify_otp(self, account: str, otp: float) -> TransactionDTO:
        raise NotImplementedError

    # Dunders

    def __repr__(self) -> str:
        return f"{self.name} / {self.currency} (preprod={self.preprod})"



# accounts
from .accounts import get_assets as get_account_assets
from .accounts import get_balance as get_account_balance
from .accounts import get_created_assets as get_account_created_assets
from .accounts import get_transactions as get_account_transactions

# assets
from .assets import get_data as get_asset_data
from .assets import get_metadata as get_asset_metadata
from .assets import get_owner as get_asset_owner

# aliases
get_collection_assets = get_account_created_assets



import json
from base64 import b64decode
from functools import lru_cache

from fetchfox.checks import check_str
from .common import get


@lru_cache(maxsize=None)
def get_data(asset_id: str) -> dict:
    check_str(asset_id, "algonodecloud.asset_id")

    response, status_code = get(
        service=f"assets/{asset_id}",
    )

    return response["asset"]["params"]


@lru_cache(maxsize=None)
def get_metadata(asset_id: str) -> dict:
    check_str(asset_id, "algonodecloud.asset_id")

    response, status_code = get(
        f"assets/{asset_id}/transactions",
        params={
            "tx-type": "acfg",
            "limit": "1",
        },
    )

    transaction = response["transactions"][0]
    note = b64decode(transaction["note"]).decode("utf-8")

    return json.loads(note)


def get_owner(asset_id: str) -> dict:
    check_str(asset_id, "algonodecloud.asset_id")

    response, status_code = get(
        service=f"assets/{asset_id}/balances",
        params={
            "currency-greater-than": "0",
        },
    )

    balances = response.get("balances", [])

    if not balances:
        return None

    return {
        "asset_id": asset_id,
        "address": balances[0]["address"],
        "amount": balances[0]["amount"],
    }



import logging
from datetime import datetime
from typing import Iterable, Tuple, List

import pytz

from fetchfox.apis import bookio
from fetchfox.apis.cardano import jpgstore, gomaestroorg
from fetchfox.blockchains.base import Blockchain
from fetchfox.constants.blockchains import CARDANO
from fetchfox.constants.currencies import ADA, BOOK
from fetchfox.constants.marketplaces import JPG_STORE
from fetchfox.constants.ranks import CNFT_RANKS
from fetchfox.dtos import (
    AssetDTO,
    CampaignDTO,
    CampaignPricingDTO,
    FloorDTO,
    HoldingDTO,
    ListingDTO,
    RankDTO,
    SaleDTO,
    SaleType,
    TransactionDTO,
    TransactionInputOutputDTO,
    TransactionAssetDTO,
)
from fetchfox.helpers import formatters
from . import utils
from .exceptions import (
    InvalidCardanoAssetIdException,
    InvalidCardanoCollectionIdException,
    InvalidCardanoAccountException,
)

WINTER_NFT_ADDRESS = "addr1qxnrv2quqxhvwxtxmygsmkufph4kjju6j5len7k2ljslpz8ql7k7gehlfvj6ektgu9ns8yx8epcp66337khxeq82rpgqe6lqyk"
SHELLY_EPOCH = 1596491091
SHELLY_SLOT = 4924800

logger = logging.getLogger(__name__)


class Cardano(Blockchain):
    def __init__(self, preprod: bool = False, gomaestroorg_api_key: str = None):
        super().__init__(
            name=CARDANO,
            currency=ADA,
            logo="https://s2.coinmarketcap.com/static/img/coins/64x64/2010.png",
            preprod=preprod,
        )

        self.gomaestroorg_api_key: str = gomaestroorg_api_key

    def check_account(self, account: str):
        if self.preprod:
            is_account = utils.is_preprod_account(account)
        else:
            is_account = utils.is_account(account)

        if not is_account:
            raise InvalidCardanoAccountException(account)

    def check_asset_id(self, asset_id: str):
        if not utils.is_asset_id(asset_id):
            raise InvalidCardanoAssetIdException(asset_id)

    def check_collection_id(self, collection_id: str):
        if not utils.is_policy_id(collection_id):
            raise InvalidCardanoCollectionIdException(collection_id)

    def explorer_url(self, *, address: str = None, collection_id: str = None, asset_id: str = None, tx_hash: str = None) -> str:
        if address:
            return f"https://pool.pm/{address.lower()}"

        if asset_id:
            return f"https://cardanoscan.io/token/{asset_id.lower()}"

        if collection_id:
            return f"https://pool.pm/policy/{collection_id.lower()}"

        if tx_hash:
            return f"https://cardanoscan.io/transaction/{tx_hash.lower()}"

        return None

    def marketplace_url(self, collection_id: str = None, asset_id: str = None) -> str:
        if asset_id:
            return f"https://www.jpg.store/asset/{asset_id.lower()}"

        if collection_id:
            return f"https://jpg.store/collection/{collection_id.lower()}"

        return None

    # Accounts

    def get_account_assets(self, wallet: str, collection_id: str = None) -> Iterable[HoldingDTO]:
        self.check_account(wallet)

        if collection_id is not None:
            self.check_collection_id(collection_id)

        stake_address = self.get_account_stake_address(wallet)

        account_assets = gomaestroorg.get_account_assets(
            stake_address,
            policy_id=collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for account_asset in account_assets:
            asset_id = account_asset["unit"]
            quantity = account_asset["amount"]

            policy_id, _, _ = utils.split_asset_id(asset_id)

            yield HoldingDTO(
                collection_id=policy_id,
                asset_id=asset_id,
                address=stake_address,
                quantity=quantity,
            )

    def get_account_name(self, account: str) -> str:
        if utils.is_ada_handle(account):
            return account

        if utils.is_address(account):
            account = self.get_account_stake_address(account)

        return gomaestroorg.get_handle(
            account,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

    def get_account_balance(self, account: str) -> Tuple[float, str]:
        stake_address = self.get_account_stake_address(account)

        balance = gomaestroorg.get_account_balance(
            stake_address,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        return balance, self.currency

    def get_account_transactions(self, account: str, last: int = 10) -> Iterable[TransactionDTO]:
        def parse_input_output(items: List[dict]) -> Iterable[TransactionInputOutputDTO]:
            for item in items:
                item_address = item["address"]
                item_assets = []

                for asset in item["assets"]:
                    item_assets.append(
                        TransactionAssetDTO(
                            amount=float(asset["amount"]),
                            unit=asset["unit"],
                        )
                    )

                yield TransactionInputOutputDTO(
                    address=item_address,
                    assets=item_assets,
                )

        self.check_account(account)

        stake_address = self.get_account_stake_address(account)

        transactions = []

        for address, tx in gomaestroorg.get_account_transactions(stake_address, api_key=self.gomaestroorg_api_key, preprod=self.preprod):
            transactions.append((address, tx))

        sorted_transactions = list(
            sorted(
                transactions,
                key=lambda x: x[1]["block_timestamp"],
                reverse=True,
            )
        )

        for address, tx in sorted_transactions[:last]:
            message = None

            if tx.get("metadata"):
                if tx["metadata"].get("674"):
                    if tx["metadata"]["674"].get("msg"):
                        message = "\n".join(tx["metadata"]["674"]["msg"])

            inputs = list(parse_input_output(tx["inputs"]))
            outputs = list(parse_input_output(tx["outputs"]))

            yield TransactionDTO(
                blockchain=self.name,
                address=stake_address,
                tx_hash=tx["tx_hash"],
                message=message,
                inputs=inputs,
                outputs=outputs,
            )

    def resolve_account_name(self, name: str) -> str:
        if not utils.is_ada_handle(name):
            return None

        resolution = gomaestroorg.resolve_handle(
            name,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        if not resolution:
            return None

        return resolution["stake_address"]

    def get_account_stake_address(self, account: str) -> str:
        self.check_account(account)

        is_stake_address = False

        if self.preprod:
            if utils.is_preprod_stake_address(account):
                is_stake_address = True
        else:
            if utils.is_stake_address(account):
                is_stake_address = True

        if is_stake_address:
            return account

        is_address = False

        if self.preprod:
            if utils.is_preprod_address(account):
                is_address = True
        else:
            if utils.is_address(account):
                is_address = True

        if is_address:
            return gomaestroorg.get_address_stake_address(
                account,
                api_key=self.gomaestroorg_api_key,
                preprod=self.preprod,
            )

        if utils.is_ada_handle(account):
            resolution = gomaestroorg.resolve_handle(
                account,
                api_key=self.gomaestroorg_api_key,
                preprod=self.preprod,
            )

            return resolution["stake_address"]

        return None

    # Assets

    def format_account(self, account: str) -> str:
        self.check_account(account)

        handle = self.get_account_name(account)

        if handle:
            punycode = handle.encode().decode("idna")

            if handle != punycode:
                return f"{punycode} (${handle})"

            return f"${handle}"

        return super().format_account(account)

    def get_asset(self, collection_id: str, asset_id: str, *args, **kwargs) -> AssetDTO:
        if collection_id is None:
            collection_id = asset_id[:56]

        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        asset_data = gomaestroorg.get_asset_data(
            asset_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        return AssetDTO(
            collection_id=collection_id,
            asset_id=asset_id,
            metadata=asset_data,
            supply=int(asset_data.get("total_supply", 1)),
        )

    def get_asset_owners(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        asset_owners = gomaestroorg.get_asset_owners(
            asset_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for owner in asset_owners:
            yield HoldingDTO(
                collection_id=collection_id,
                asset_id=owner["asset_id"],
                address=owner["stake_address"],
                quantity=owner["amount"],
            )

    def get_asset_rank(self, collection_id: str, asset_id: str, *args, **kwargs) -> RankDTO:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        _, _, decoded_asset_name = utils.split_asset_id(asset_id)

        if not decoded_asset_name:
            return None

        number = int("".join((c for c in decoded_asset_name if c.isdigit())))

        if collection_id not in CNFT_RANKS:
            return None

        if decoded_asset_name not in CNFT_RANKS[collection_id]:
            return None

        return RankDTO(
            collection_id=collection_id,
            number=number,
            asset_id=asset_id,
            rank=CNFT_RANKS[collection_id][decoded_asset_name],
        )

    # Collections

    def get_collection_assets(
        self,
        collection_id: str,
        discriminator: str = None,
        fetch_metadata: bool = True,
        *args,
        **kwargs,
    ) -> Iterable[AssetDTO]:
        self.check_collection_id(collection_id)

        if discriminator:
            discriminator = discriminator.lower()

        collection_assets = gomaestroorg.get_policy_assets(
            collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for asset in collection_assets:
            asset_name = asset["asset_name"].lower()

            if not asset_name:  # skip royalty token
                continue

            asset_id = f"{collection_id}{asset_name}"
            cip25_metadata = asset["asset_standards"].get("cip25_metadata")
            cip68_metadata = asset["asset_standards"].get("cip68_metadata")

            # required for multi-book policies (e.g. monsters, greek classics)
            if discriminator:
                if not asset_name.startswith(discriminator):
                    continue

            yield AssetDTO(
                collection_id=collection_id,
                asset_id=asset_id,
                metadata=cip68_metadata or cip25_metadata,
                supply=int(asset.get("total_supply", 1)),
            )

    def get_collection_lock_date(self, collection_id: str, *args, **kwargs) -> datetime:
        slot = gomaestroorg.get_policy_lock_slot(
            collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        if slot is None:
            return None

        epoch = SHELLY_EPOCH + (slot - SHELLY_SLOT)
        return datetime.utcfromtimestamp(epoch).replace(tzinfo=pytz.UTC)

    def get_collection_floor(self, collection_id: str, discriminator: str = None, *args, **kwargs) -> FloorDTO:
        self.check_collection_id(collection_id)

        collection_listings = self.get_collection_listings(
            collection_id,
            discriminator=discriminator,
        )

        floor = None
        count = 0

        for listing in collection_listings:
            count += 1

            if floor is None:
                floor = listing
            elif listing.usd < floor.usd:
                floor = listing

        return FloorDTO(
            listing=floor,
            listing_count=count,
        )

    def get_collection_listings(self, collection_id: str, discriminator: str = None, *args, **kwargs) -> Iterable[ListingDTO]:
        self.check_collection_id(collection_id)

        if discriminator:
            discriminator = discriminator.lower()

        collection_listings = jpgstore.get_collection_listings(collection_id)

        for listing in collection_listings:
            asset_id = listing["asset_id"]
            policy_id, asset_name, _ = utils.split_asset_id(asset_id)

            # required for multi-book policies (e.g. monsters, greek classics)
            if discriminator:
                if not asset_name.startswith(discriminator):
                    continue

            asset_ids = []
            asset_names = []

            if listing["listing_type"] == "BUNDLE":
                for bundled_asset in listing["bundled_assets"]:
                    asset_ids.append(bundled_asset["asset_id"])
                    asset_names.append(bundled_asset["display_name"])
            else:
                asset_ids.append(listing["asset_id"])
                asset_names.append(listing["display_name"])

            if listing.get("confirmed_at"):
                listed_at = datetime.fromisoformat(listing["confirmed_at"].replace("Z", "+00:00"))
            else:
                listed_at = datetime.now(tz=pytz.utc)

            yield ListingDTO(
                identifier=listing["tx_hash"],
                collection_id=policy_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                listing_id=listing["listing_id"],
                marketplace=JPG_STORE,
                price=int(listing["price_lovelace"]) // 10**6,
                currency=self.currency,
                listed_at=listed_at,
                listed_by=None,
                tx_hash=listing["tx_hash"],
                marketplace_url=f"https://jpg.store/asset/{asset_id}",
            )

    def get_collection_ranks(self, collection_id: str, *args, **kwargs) -> Iterable[RankDTO]:
        self.check_collection_id(collection_id)

        if collection_id not in CNFT_RANKS:
            return

        for asset_name, rank in CNFT_RANKS[collection_id].items():
            number = int("".join((c for c in asset_name if c.isdigit())))

            yield RankDTO(
                collection_id=collection_id,
                number=number,
                asset_id=None,
                rank=int(rank),
            )

    def get_collection_sales(self, collection_id: str, discriminator: str = None, *args, **kwargs) -> Iterable[SaleDTO]:
        self.check_collection_id(collection_id)

        if discriminator:
            discriminator = discriminator.lower()

        collection_sales = jpgstore.get_collection_sales(collection_id)

        for sale in collection_sales:
            tx_hash = sale["tx_hash"]

            asset_id = sale["asset_id"]
            policy_id, asset_name, encoded_asset_name = utils.split_asset_id(asset_id)

            # required for multi-book policies (e.g. monsters, greek classics)
            if discriminator:
                if not encoded_asset_name.startswith(discriminator):
                    continue

            bulk_size = sale.get("bulk_size", None) or 1

            if sale["action"] == "ACCEPT_OFFER":
                buyer = sale["seller_address"]
                seller = sale["signer_address"]
                sale_type = SaleType.OFFER
            elif sale["action"] == "ACCEPT_COLLECTION_OFFER":
                buyer = sale["signer_address"]
                seller = sale["seller_address"]
                sale_type = SaleType.COLLECTION_OFFER
            elif sale["action"] == "BUY":
                buyer = sale["signer_address"]
                seller = sale["seller_address"]

                if buyer == WINTER_NFT_ADDRESS:
                    sale_type = SaleType.CREDIT_CARD
                else:
                    sale_type = SaleType.PURCHASE
            else:
                continue

            asset_ids = []
            asset_names = []

            if sale["listing_from_tx_history"]["bundled_assets"]:
                for bundled_asset in sale["listing_from_tx_history"]["bundled_assets"]:
                    asset_ids.append(bundled_asset["asset_id"])
                    asset_names.append(bundled_asset["display_name"])
            else:
                asset_ids.append(sale["asset_id"])
                asset_names.append(sale["display_name"])

            if sale.get("confirmed_at"):
                confirmed_at = datetime.fromisoformat(sale["confirmed_at"].replace("Z", "+00:00"))
            else:
                confirmed_at = datetime.now(tz=pytz.utc)

            yield SaleDTO(
                identifier=tx_hash,
                collection_id=policy_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                tx_hash=tx_hash,
                marketplace=JPG_STORE,
                price=int(sale["amount_lovelace"]) // 10**6,
                currency=self.currency,
                confirmed_at=confirmed_at,
                type=sale_type,
                bulk_size=bulk_size,
                sold_by=seller,
                bought_by=buyer,
                marketplace_url=f"https://jpg.store/asset/{asset_id}",
                explorer_url=f"https://cardanoscan.io/transaction/{tx_hash}",
            )

    def get_collection_snapshot(self, collection_id: str, discriminator: str = None, max_threads: int = 3, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)

        collection_owners = gomaestroorg.get_policy_owners(
            collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for owner in collection_owners:
            stake_address = owner["account"]

            for asset in owner.get("assets", []):
                encoded_asset_name = asset["name"].lower()
                asset_id = f"{collection_id}{encoded_asset_name}"
                quantity = asset["amount"]

                # required for multi-book policies (e.g. monsters, greek classics)
                if discriminator:
                    if not encoded_asset_name.startswith(discriminator.lower()):
                        continue

                yield HoldingDTO(
                    collection_id=collection_id,
                    asset_id=asset_id,
                    address=stake_address,
                    quantity=quantity,
                )

    # Others

    def verify_otp(self, account: str, otp: int) -> TransactionDTO:
        for transaction in self.get_account_transactions(account):
            invalid = True

            # check if account is in at least one input
            for tx_input in transaction.inputs:
                if tx_input.address == account:
                    invalid = False
                    break

            if invalid:
                continue

            # check if account is in at least one output with the correct amount
            for tx_output in transaction.outputs:
                if tx_output.address != account:
                    continue

                for tx_asset in tx_output.assets:
                    if tx_asset.unit != "lovelace":
                        continue

                    if tx_asset.amount != otp:
                        continue

                    return transaction

        return None

    # Campaigns

    def get_campaigns(self, starts_after: datetime = None) -> Iterable[CampaignDTO]:
        def parse_pricing(pricing: dict) -> List[CampaignPricingDTO]:
            result = []

            if pricing.get("native_price"):
                result.append(
                    CampaignPricingDTO(
                        currency=ADA,
                        amount=pricing["native_price"] / 10**6,
                    )
                )

            if pricing.get("book_token_price"):
                result.append(
                    CampaignPricingDTO(
                        currency=BOOK,
                        amount=pricing["book_token_price"] / 10**6,
                    )
                )

            return result

        for campaign in bookio.get_campaigns():
            if not campaign.get("collection_id"):
                continue

            if campaign["blockchain"] != "cardano":
                continue

            if campaign["network"] != "mainnet":
                continue

            collection_id = campaign["collection_id"]
            start_at = formatters.timestamp(campaign["start_at"])

            if starts_after is not None:
                if start_at < starts_after:
                    continue

            yield CampaignDTO(
                blockchain=self.name,
                parlamint_id=campaign["campaign_id"],
                collection_id=collection_id,
                name=campaign["name"],
                start_at=start_at,
                supply=campaign["total_deas"],
                limit=campaign["max_quantity"],
                mint_pricing=parse_pricing(
                    pricing=campaign.get("mint_price", {}),
                ),
                discount_pricing=parse_pricing(
                    pricing=campaign.get("discount_price", {}),
                ),
                bookstore_url=campaign["bookstore_url"],
                cover_url=campaign["cover_url"],
                explorer_url=self.explorer_url(
                    collection_id=collection_id,
                ),
            )



# addresses
from .addresses import get_assets as get_addresses_assets
from .addresses import get_balance as get_addresses_balance

# assets
from .assets import get_data as get_asset_data
from .assets import get_owner as get_asset_owner

# contracts
from .contracts import get_assets as get_contract_assets
from .contracts import get_owners as get_contract_owners

# wallets
from .wallets import get_transactions as get_wallet_transactions

# aliases
get_account_assets = get_addresses_assets
get_account_balance = get_addresses_balance
get_account_transactions = get_wallet_transactions
get_collection_assets = get_contract_assets
get_collection_owners = get_contract_owners



# collections
from .collections import get_sales as get_collection_sales

# policies
from .policies import get_listings as get_policy_listings

# aliases
get_collection_listings = get_policy_listings
